from typing import Callable, Dict, IO, Iterable, Optional, Union
from .io.binary_encoder import BinaryEncoder
from .types import AvroMessage


def writer(
        fo: IO,
        schema: Dict,
        records: Iterable,
        codec: str,
        sync_interval: int,
        metadata: Optional[Dict],
        validator: Union[Callable, bool, None],
        sync_marker: Optional[bytes],
        codec_compression_level: Optional[int]) -> None: ...

class GenericWriter:
    schema: Dict
    validate_fn: Callable
    metadata: Dict


class Writer(GenericWriter):
    encoder: BinaryEncoder
    io: BinaryEncoder
    block_count: int
    sync_interval: int
    compression_level: Optional[int]
    block_writer: Callable

    def __init__(
            self,
            fo: IO,
            records: Iterable,
            codec: str,
            sync_interval: int,
            metadata: Optional[Dict],
            validator: Union[Callable, bool, None],
            sync_marker: Optional[bytes],
            codec_compression_level: Optional[int]): ...

    def dump(self) -> None: ...
    def write(self, record: AvroMessage) -> None: ...
    def write_block(self, block) -> None: ...
    def flush(self) -> None: ...


def schemaless_writer(fo: IO, schema: Dict, record: Dict) -> None: ...



import fastavro
from fastavro.__main__ import _clean_json_record
from fastavro._timezone import epoch
import pytest

from decimal import Decimal
from io import BytesIO
from uuid import uuid4
import datetime
import sys
import os
from dateutil.tz import tzlocal

from .conftest import assert_naive_datetime_equal_to_tz_datetime


schema = {
    "fields": [
        {
            "name": "date",
            "type": {'type': 'int', 'logicalType': 'date'}
        },
        {
            "name": "timestamp-millis",
            "type": {'type': 'long', 'logicalType': 'timestamp-millis'}
        },
        {
            "name": "timestamp-micros",
            "type": {'type': 'long', 'logicalType': 'timestamp-micros'}
        },
        {
            "name": "uuid",
            "type": {'type': 'string', 'logicalType': 'uuid'}
        },
        {
            "name": "time-millis",
            "type": {'type': 'int', 'logicalType': 'time-millis'}
        },
        {
            "name": "time-micros",
            "type": {'type': 'long', 'logicalType': 'time-micros'}
        }
    ],
    "namespace": "namespace",
    "name": "name",
    "type": "record"
}


def serialize(schema, data):
    bytes_writer = BytesIO()
    fastavro.schemaless_writer(bytes_writer, schema, data)
    return bytes_writer.getvalue()


def deserialize(schema, binary):
    bytes_writer = BytesIO()
    bytes_writer.write(binary)
    bytes_writer.seek(0)

    res = fastavro.schemaless_reader(bytes_writer, schema)
    return res


def test_logical_types():
    data1 = {
        'date': datetime.date.today(),
        'timestamp-millis': datetime.datetime.now(),
        'timestamp-micros': datetime.datetime.now(),
        'uuid': uuid4(),
        'time-millis': datetime.datetime.now().time(),
        'time-micros': datetime.datetime.now().time(),

    }
    binary = serialize(schema, data1)
    data2 = deserialize(schema, binary)
    assert (data1['date'] == data2['date'])
    assert_naive_datetime_equal_to_tz_datetime(
        data1['timestamp-micros'],
        data2['timestamp-micros'],
    )
    assert (int(data1['timestamp-millis'].microsecond / 1000) * 1000
            == data2['timestamp-millis'].microsecond)
    assert (int(data1['time-millis'].microsecond / 1000) * 1000
            == data2['time-millis'].microsecond)


@pytest.mark.skipif(os.name == 'nt' and sys.version_info[:2] == (3, 6),
                    reason='Python Bug: https://bugs.python.org/issue29097')
def test_not_logical_ints():
    data1 = {
        'date': 1,
        'timestamp-millis': 2,
        'timestamp-micros': 3,
        'uuid': uuid4(),
        'time-millis': 4,
        'time-micros': 5,

    }
    binary = serialize(schema, data1)
    data2 = deserialize(schema, binary)
    # 1 day since 1970-1-1
    assert (data2['date'] == datetime.date(1970, 1, 2))


schema_null = {
    "fields": [
        {
            "name": "date",
            "type": ["null", {'type': 'int', 'logicalType': 'date'}]
        },
    ],
    "namespace": "namespace",
    "name": "name",
    "type": "record"
}


def test_null():
    data1 = {
        # 'date': None,
    }
    binary = serialize(schema_null, data1)
    data2 = deserialize(schema_null, binary)
    assert (data2['date'] is None)


def test_not_null_datetime():
    data1 = {'date': datetime.datetime(2017, 1, 1)}
    binary = serialize(schema_null, data1)
    data2 = deserialize(schema_null, binary)
    assert (data2['date'] == datetime.date(2017, 1, 1))


def test_not_null_date():
    data1 = {'date': datetime.date(2017, 1, 1)}
    binary = serialize(schema_null, data1)
    data2 = deserialize(schema_null, binary)
    assert (data2['date'] == datetime.date(2017, 1, 1))


# particularly critical on Windows
# see https://github.com/fastavro/fastavro/issues/389
@pytest.mark.skipif(
    os.name == "nt" and sys.version_info <= (3, 0),
    reason="Bug was not fixed on Python 2"
)
def test_ancient_datetime():
    schema_datetime = {
        "fields": [
            {
                "name": "timestamp-millis",
                "type": {'type': 'long', 'logicalType': 'timestamp-millis'}
            },
            {
                "name": "timestamp-micros",
                "type": {'type': 'long', 'logicalType': 'timestamp-micros'}
            }
        ],
        "namespace": "namespace",
        "name": "name",
        "type": "record"
    }

    my_epoch = datetime.datetime(1970, 1, 1, tzinfo=tzlocal())
    diff = my_epoch - epoch

    data1 = {
        'timestamp-millis': datetime.datetime(1960, 1, 1),
        'timestamp-micros': datetime.datetime(1960, 1, 1)
    }
    binary = serialize(schema_datetime, data1)
    data2 = deserialize(schema_datetime, binary)

    assert (
        (data1['timestamp-millis'] + diff)
        == data2['timestamp-millis'].replace(tzinfo=None)
    )
    assert (
        (data1['timestamp-micros'] + diff)
        == data2['timestamp-micros'].replace(tzinfo=None)
    )


# test bytes decimal
schema_bytes_decimal = {
    "name": "n",
    "namespace": "namespace",
    "type": "bytes",
    "logicalType": "decimal",
    "precision": 20,
    "scale": 3,
}


@pytest.mark.parametrize(
    'input_data, expected_binary',
    [
        (Decimal("0.0"), b'\x02\x00'),
        (Decimal("-0.0"), b'\x02\x00'),
        (Decimal("0.1"), b'\x02d'),
        (Decimal("-0.1"), b'\x02\x9c'),
        (Decimal("0.2"), b'\x04\x00\xc8'),
        (Decimal("-0.2"), b'\x04\xff8'),
        (Decimal("0.456"), b'\x04\x01\xc8'),
        (Decimal("-0.456"), b'\x04\xfe8'),
        (Decimal("2.55"), b'\x04\t\xf6'),
        (Decimal("-2.55"), b'\x04\xf6\n'),
        (Decimal("2.90"), b'\x04\x0bT'),
        (Decimal("-2.90"), b'\x04\xf4\xac'),
        (Decimal("123.456"), b'\x06\x01\xe2@'),
        (Decimal("-123.456"), b'\x06\xfe\x1d\xc0'),
        (Decimal("3245.234"), b'\x061\x84\xb2'),
        (Decimal("-3245.234"), b'\x06\xce{N'),
        (
            Decimal("9999999999999999.456"),
            b'\x12\x00\x8a\xc7#\x04\x89\xe7\xfd\xe0'
        ),
        (Decimal("-999999999999999.456"), b'\x10\xf2\x1fILX\x9c\x02 '),
    ]
)
def test_bytes_decimal(input_data, expected_binary):
    binary = serialize(schema_bytes_decimal, input_data)
    assert (binary == expected_binary)
    output_data = deserialize(schema_bytes_decimal, binary)
    assert (input_data == output_data)


def test_bytes_decimal_scale():
    data1 = Decimal("123.456678")  # does not fit scale
    with pytest.raises(ValueError):
        serialize(schema_bytes_decimal, data1)


def test_bytes_decimal_precision():
    data1 = Decimal("123456789012345678.901")  # does not fit precision
    with pytest.raises(ValueError):
        serialize(schema_bytes_decimal, data1)


schema_bytes_decimal_leftmost = {
    "name": "n",
    "namespace": "namespace",
    "type": "bytes",
    "logicalType": "decimal",
    "precision": 18,
    "scale": 6,
}


def test_bytes_decimal_leftmost():
    binary = serialize(schema_bytes_decimal_leftmost, b'\xd5F\x80')
    data2 = deserialize(schema_bytes_decimal_leftmost, binary)
    assert (Decimal("-2.80") == data2)


# test fixed decimal
schema_fixed_decimal = {
    "name": "n",
    "namespace": "namespace",
    "type": "fixed",
    "size": 8,
    "logicalType": "decimal",
    "precision": 15,
    "scale": 3,
}


def test_fixed_decimal_negative():
    data1 = Decimal("-2.90")
    binary = serialize(schema_fixed_decimal, data1)
    data2 = deserialize(schema_fixed_decimal, binary)
    assert (data1 == data2)
    assert (len(binary) == schema_fixed_decimal['size'])


def test_fixed_decimal_zero():
    data1 = Decimal("0.0")
    binary = serialize(schema_fixed_decimal, data1)
    data2 = deserialize(schema_fixed_decimal, binary)
    assert (data1 == data2)
    assert (len(binary) == schema_fixed_decimal['size'])


def test_fixed_decimal_positive():
    data1 = Decimal("123.456")
    binary = serialize(schema_fixed_decimal, data1)
    data2 = deserialize(schema_fixed_decimal, binary)
    assert (data1 == data2)
    assert (len(binary) == schema_fixed_decimal['size'])


def test_fixed_decimal_scale():
    data1 = Decimal("123.456678")  # does not fit scale
    with pytest.raises(ValueError):
        serialize(schema_fixed_decimal, data1)


def test_fixed_decimal_precision():
    data1 = Decimal("123456789012345678.901")  # does not fit precision
    with pytest.raises(ValueError):
        serialize(schema_fixed_decimal, data1)


schema_fixed_decimal_leftmost = {
    "name": "n",
    "namespace": "namespace",
    "type": "fixed",
    "size": 8,
    "logicalType": "decimal",
    "precision": 18,
    "scale": 6,
}


def test_fixed_decimal_binary():
    binary = serialize(schema_fixed_decimal_leftmost,
                       b'\xFF\xFF\xFF\xFF\xFF\xd5F\x80')
    data2 = deserialize(schema_fixed_decimal_leftmost, binary)
    assert (Decimal("-2.80") == data2)


def test_clean_json_list():
    values = [
        datetime.datetime.now(),
        datetime.date.today(),
        uuid4(),
        Decimal('1.23'),
    ]
    str_values = [
        values[0].isoformat(),
        values[1].isoformat(),
        str(values[2]),
        str(values[3]),
    ]
    _clean_json_record(values)
    assert values == str_values


def test_clean_json_dict():
    values = {
        '1': datetime.datetime.now(),
        '2': datetime.date.today(),
        '3': uuid4(),
        '4': Decimal('1.23'),
    }
    str_values = {
        '1': values['1'].isoformat(),
        '2': values['2'].isoformat(),
        '3': str(values['3']),
        '4': str(values['4']),
    }
    _clean_json_record(values)
    assert values == str_values


def test_unknown_logical_type():
    unknown_type_schema = {
        "type": "record",
        "name": "t_customers",
        "fields": [{
            "name": "name",
            "type": {
                "type": "string",
                "logicalType": "varchar",
                "maxLength": 200
            },
        }, {
            "name": "address",
            "type": {
                "type": "record",
                "name": "t_address",
                "fields": [{
                    "name": "street",
                    "type": {
                        "type": "string",
                        "logicalType": "varchar",
                        "maxLength": 240
                    }
                }, {
                    "name": "city",
                    "type": {
                        "type": "string",
                        "logicalType": "varchar",
                        "maxLength": 80
                    }
                }, {
                    "name": "zip",
                    "type": {
                        "type": "string",
                        "logicalType": "varchar",
                        "maxLength": 18
                    }
                }]
            }
        }]
    }

    data1 = {
        'name': 'foo',
        'address': {
            'street': '123 street',
            'city': 'city',
            'zip': '00000'
        }
    }
    converted = serialize(unknown_type_schema, data1)
    data2 = deserialize(unknown_type_schema, converted)
    assert (data1 == data2)


def test_default_scale_value():
    schema = {
        "name": "test_default_scale_value",
        "type": "fixed",
        "size": 8,
        "logicalType": "decimal",
        "precision": 15,
    }

    data1 = Decimal("-2")
    binary = serialize(schema, data1)
    data2 = deserialize(schema, binary)
    assert (data1 == data2)


def test_date_as_string():
    schema = {
        "name": "test_date_as_string",
        "type": "int",
        "logicalType": "date",
    }

    data1 = "2019-05-06"
    binary = serialize(schema, data1)
    data2 = deserialize(schema, binary)
    assert (datetime.date(2019, 5, 6) == data2)


@pytest.mark.skipif(
    hasattr(sys, 'pypy_version_info'),
    reason='pandas takes forever to install on pypy'
)
def test_pandas_datetime():
    """https://github.com/gojek/feast/pull/490#issuecomment-590623525"""

    # Import here as pandas is not installed on pypy for testing
    import pandas as pd
    import pytz

    schema = {
        "fields": [
            {
                "name": "timestamp-micros",
                "type": [
                    "null", {'type': 'long', 'logicalType': 'timestamp-micros'}
                ]
            }
        ],
        "name": "test_pandas_datetime",
        "type": "record",
    }

    data1 = {
        "timestamp-micros": pd.to_datetime(
            datetime.datetime.utcnow().replace(tzinfo=pytz.utc)
        )
    }
    assert serialize(schema, data1)



# cython: language_level=3str
# cython: auto_cpdef=True

from os import path
from copy import deepcopy
import json
from libc.math cimport floor, log10

from ._six import iteritems
from ._schema_common import (
    PRIMITIVES, UnknownType, SchemaParseException, RESERVED_PROPERTIES,
    OPTIONAL_FIELD_PROPERTIES, RESERVED_FIELD_PROPERTIES,
)


cpdef inline extract_record_type(schema):
    if isinstance(schema, dict):
        return schema['type']

    if isinstance(schema, list):
        return 'union'

    return schema


cpdef inline str extract_logical_type(schema):
    if not isinstance(schema, dict):
        return None
    rt = schema['type']
    lt = schema.get('logicalType')
    if lt:
        # TODO: Building this string every time is going to be relatively slow.
        return '{}-{}'.format(rt, lt)
    return None


cpdef fullname(schema):
    return schema_name(schema, "")[1]


cpdef schema_name(schema, parent_ns):
    try:
        name = schema['name']
    except KeyError:
        msg = (
            '"name" is a required field missing from '
            + 'the schema: {}'.format(schema)
        )
        raise SchemaParseException(msg)

    namespace = schema.get('namespace', parent_ns)
    if not namespace:
        return namespace, name

    return namespace, '{}.{}'.format(namespace, name)


cpdef expand_schema(schema):
    return parse_schema(schema, expand=True, _write_hint=False)


def parse_schema(
    schema, expand=False, _write_hint=True, _force=False, _named_schemas=None
):
    if _named_schemas is None:
        _named_schemas = {}

    if _force or expand:
        return _parse_schema(
            schema, "", expand, _write_hint, set(), _named_schemas
        )
    elif isinstance(schema, dict) and "__fastavro_parsed" in schema:
        for key, value in iteritems(schema["__named_schemas"]):
            _named_schemas[key] = value
        return schema
    elif isinstance(schema, list):
        # If we are given a list we should make sure that the immediate sub
        # schemas have the hint in them
        return [
            parse_schema(s, expand, _write_hint, _force, _named_schemas)
            for s in schema
        ]
    else:
        return _parse_schema(
            schema, "", expand, _write_hint, set(), _named_schemas
        )


cdef _parse_schema(schema, namespace, expand, _write_hint, names, named_schemas):
    # union schemas
    if isinstance(schema, list):
        return [
            _parse_schema(
                s, namespace, expand, False, names, named_schemas
            ) for s in schema
        ]

    # string schemas; this could be either a named schema or a primitive type
    elif not isinstance(schema, dict):
        if schema in PRIMITIVES:
            return schema

        if '.' not in schema and namespace:
            schema = namespace + '.' + schema

        if schema not in named_schemas:
            raise UnknownType(schema)
        elif expand:
            # If `name` is in the schema, it has been fully resolved and so we
            # can include the full schema. If `name` is not in the schema yet,
            # then we are still recursing that schema and must use the named
            # schema or else we will have infinite recursion when printing the
            # final schema
            if "name" in named_schemas[schema]:
                return named_schemas[schema]
            else:
                return schema
        else:
            return schema

    else:
        # Remaining valid schemas must be dict types
        schema_type = schema["type"]

        parsed_schema = {
            key: value
            for key, value in iteritems(schema)
            if key not in RESERVED_PROPERTIES
        }
        parsed_schema["type"] = schema_type

        if "doc" in schema:
            parsed_schema["doc"] = schema["doc"]

        # Correctness checks for logical types
        logical_type = parsed_schema.get("logicalType")
        if logical_type == "decimal":
            scale = parsed_schema.get("scale")
            if scale and (not isinstance(scale, int) or scale < 0):
                raise SchemaParseException(
                    "decimal scale must be a postive integer, "
                    + "not {}".format(scale)
                )

            precision = parsed_schema.get("precision")
            if precision:
                if not isinstance(precision, int) or precision <= 0:
                    raise SchemaParseException(
                        "decimal precision must be a postive integer, "
                        + "not {}".format(precision)
                    )
                if schema_type == "fixed":
                    # https://avro.apache.org/docs/current/spec.html#Decimal
                    size = schema["size"]
                    max_precision = int(floor(log10(2) * (8 * size - 1)))
                    if precision > max_precision:
                        msg = "decimal precision of {} doesn't fit into " \
                              "array of length {}"
                        raise SchemaParseException(
                            msg.format(precision, size)
                        )

            if scale and precision and precision < scale:
                raise SchemaParseException(
                    "decimal scale must be less than or equal to "
                    + "the precision of {}".format(precision)
                )

        if schema_type == "array":
            parsed_schema["items"] = _parse_schema(
                schema["items"],
                namespace,
                expand,
                False,
                names,
                named_schemas,
            )

        elif schema_type == "map":
            parsed_schema["values"] = _parse_schema(
                schema["values"],
                namespace,
                expand,
                False,
                names,
                named_schemas,
            )

        elif schema_type == "enum":
            _, fullname = schema_name(schema, namespace)
            if fullname in names:
                raise SchemaParseException(
                    "redefined named type: {}".format(fullname)
                )
            names.add(fullname)

            named_schemas[fullname] = parsed_schema

            parsed_schema["name"] = fullname
            parsed_schema["symbols"] = schema["symbols"]

        elif schema_type == "fixed":
            _, fullname = schema_name(schema, namespace)
            if fullname in names:
                raise SchemaParseException(
                    "redefined named type: {}".format(fullname)
                )
            names.add(fullname)

            named_schemas[fullname] = parsed_schema

            parsed_schema["name"] = fullname
            parsed_schema["size"] = schema["size"]

        elif schema_type == "record" or schema_type == "error":
            # records
            namespace, fullname = schema_name(schema, namespace)
            if fullname in names:
                raise SchemaParseException(
                    "redefined named type: {}".format(fullname)
                )
            names.add(fullname)

            named_schemas[fullname] = parsed_schema

            fields = []
            for field in schema.get('fields', []):
                fields.append(
                    parse_field(field, namespace, expand, names, named_schemas)
                )

            parsed_schema["name"] = fullname
            parsed_schema["fields"] = fields

            # Hint that we have parsed the record
            if _write_hint:
                # Make a copy of parsed_schema so that we don't have a cyclical
                # reference. Using deepcopy is pretty slow, and we don't need a
                # true deepcopy so this works good enough
                named_schemas[fullname] = {
                    k: v for k, v in iteritems(parsed_schema)
                }

                parsed_schema["__fastavro_parsed"] = True
                parsed_schema["__named_schemas"] = named_schemas

        elif schema_type in PRIMITIVES:
            parsed_schema["type"] = schema_type

        else:
            raise UnknownType(schema)

        return parsed_schema


cdef parse_field(field, namespace, expand, names, named_schemas):
    parsed_field = {
        key: value
        for key, value in iteritems(field)
        if key not in RESERVED_FIELD_PROPERTIES
    }

    for prop in OPTIONAL_FIELD_PROPERTIES:
        if prop in field:
            parsed_field[prop] = field[prop]

    # Aliases must be a list
    aliases = parsed_field.get("aliases", [])
    if not isinstance(aliases, list):
        msg = "aliases must be a list, not {}".format(aliases)
        raise SchemaParseException(msg)

    parsed_field["name"] = field["name"]
    parsed_field["type"] = _parse_schema(
        field["type"], namespace, expand, False, names, named_schemas
    )

    return parsed_field


def load_schema(schema_path, _named_schemas=None):
    '''
    Returns a schema loaded from the file at `schema_path`.

    Will recursively load referenced schemas assuming they can be found in
    files in the same directory and named with the convention
    `<type_name>.avsc`.
    '''
    if _named_schemas is None:
        _named_schemas = {}

    with open(schema_path) as fd:
        schema = json.load(fd)
    schema_dir, schema_file = path.split(schema_path)
    return _load_schema(schema, schema_dir, _named_schemas)


cdef _load_schema(schema, schema_dir, named_schemas):
    try:
        schema_copy = deepcopy(named_schemas)
        return parse_schema(schema, _named_schemas=named_schemas)
    except UnknownType as e:
        try:
            avsc = path.join(schema_dir, '%s.avsc' % e.name)
            sub_schema = load_schema(avsc)
        except IOError:
            raise e

        if isinstance(schema, dict):
            if isinstance(sub_schema, list):
                return _load_schema(
                    sub_schema + [schema], schema_dir, schema_copy
                )
            else:
                return _load_schema(
                    [sub_schema, schema], schema_dir, schema_copy
                )
        else:
            # schema is already a list
            if isinstance(sub_schema, list):
                return _load_schema(
                    sub_schema + schema, schema_dir, schema_copy
                )
            else:
                schema.insert(0, sub_schema)
                return _load_schema(schema, schema_dir, schema_copy)



import decimal
from typing import Any, Callable, Dict, Iterator, IO, Optional, Union, Tuple
from .types import AvroMessage


class reader:
    fo: IO
    return_record_name: bool
    metadata: Dict[str, bytes]
    codec: str
    reader_schema: Optional[Dict]
    writer_schema: Optional[Dict]
    def __init__(self,
                 fo: IO,
                 reader_schema: Optional[Dict],
                 return_record_name: bool): ...
    def __iter__(self) -> Iterator[AvroMessage]: ...
    def next(self) -> AvroMessage: ...
    def __next__(self) -> AvroMessage: ...


class block_reader:
    fo: IO
    return_record_name: bool
    metadata: Dict[str, bytes]
    codec: str
    reader_schema: Optional[Dict]
    writer_schema: Optional[Dict]
    def __init__(self,
                 fo: IO,
                 reader_schema: Optional[Dict],
                 return_record_name: bool): ...
    def __iter__(self) -> Iterator[Block]: ...
    def next(self) -> Block: ...
    def __next__(self) -> Block: ...


class Block:
    num_records: int
    writer_schema: Dict
    reader_schema: Dict
    offset: int
    size: int
    def __init__(
            self,
            bytes_: bytes,
            num_records: int,
            codec: str,
            reader_schema: Dict,
            writer_schema: Dict,
            offset: int,
            size: int,
            return_record_name: bool): ...
    def __iter__(self) -> Iterator[AvroMessage]: ...
    def __str__(self) -> str: ...


def json_reader(fo: IO, schema: Dict) -> reader: ...


def schemaless_reader(
        fo: IO,
        writer_schema: Dict,
        reader_schema: Optional[Dict],
        return_record_name: bool) -> AvroMessage: ...


def is_avro(path_or_buffer: Union[str, IO]) -> bool: ...


logical_reader = Callable[[Any, Optional[Dict], Optional[Dict]], Any]
LOGICAL_READERS: Dict[str, logical_reader]


class SchemaResolutionError(Exception): ...


BLOCK_READERS: Dict[str, Callable]



import decimal
from typing import Union, List, Dict

AvroMessage = Union[
    None,            # 'null' Avro type
    str,             # 'string' and 'enum'
    float,           # 'float' and 'double'
    int,             # 'int' and 'long'
    decimal.Decimal, # 'fixed'
    bool,            # 'boolean'
    bytes,           # 'bytes'
    List,            # 'array'
    Dict,            # 'map' and 'record'
]
